/*
 * Copyright (C) 2008, Uwe Schmidt 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions: 
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE. 
 */
package org.uweschmidt.dijkstravis.test;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertNotSame;
import static junit.framework.Assert.assertTrue;

import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import org.junit.Test;
import org.uweschmidt.dijkstravis.graph.IEdge;
import org.uweschmidt.dijkstravis.graph.IGraph;
import org.uweschmidt.dijkstravis.graph.INode;
import org.uweschmidt.dijkstravis.graph.algorithm.AlgorithmProgressListener;
import org.uweschmidt.dijkstravis.graph.algorithm.DijkstraShortestPath;
import org.uweschmidt.dijkstravis.graph.algorithm.DijkstraStepChanges;
import org.uweschmidt.dijkstravis.graph.jung.MyEdgeFactory;
import org.uweschmidt.dijkstravis.graph.jung.MyGraphFactory;
import org.uweschmidt.dijkstravis.graph.jung.MyNodeFactory;
import org.uweschmidt.dijkstravis.graph.jung.MyPseudoPlanarGraphGenerator;


public class AlgorithmTest {

	private int step = 0;
	private DijkstraStepChanges c;

	@Test
	public void testNodeAndEdgeAttributeConsistency() {
		MyNodeFactory.getInstance().reset();
		MyEdgeFactory.getInstance().reset();
		final MyPseudoPlanarGraphGenerator<INode, IEdge> gen = new MyPseudoPlanarGraphGenerator<INode, IEdge>(MyGraphFactory.getInstance(), MyNodeFactory.getInstance(),
				MyEdgeFactory.getInstance());
		final IGraph graph = (IGraph) gen.generateGraph(10, 10, .8, .5);

		final DijkstraShortestPath algo = new DijkstraShortestPath();
		algo.addProgressListener(new AlgorithmProgressListener<DijkstraStepChanges>() {
			public void initialized(int maxSteps) {
			}

			public void reset() {
			}

			public void stepChanged(int step, DijkstraStepChanges changes) {
				AlgorithmTest.this.step = step;
				AlgorithmTest.this.c = changes;
			}
		});
		algo.setGraph(graph);

		final Map<INode, INode.Attribute> va = new HashMap<INode, INode.Attribute>();
		final Map<IEdge, IEdge.Attribute> ea = new HashMap<IEdge, IEdge.Attribute>();

		final INode start = graph.getNodes().iterator().next();
		algo.initialize(start);

		final Random rand = new Random();
		for (INode v : graph.getNodes()) {
			va.put(v, v.getAttribute());
			for (IEdge e : graph.getOutEdges(v)) {
				e.setLength(rand.nextDouble()*100);
				ea.put(e, e.getAttribute());
			}
		}

		while (!algo.isFinished()) {
			algo.stepForward();
			checkAttributes(graph);
			algo.stepBackward();
			checkAttributePreservation(graph, va, ea);
			algo.stepForward();
			checkAttributes(graph);
			updateAttributes(graph, va, ea);
		}
	}

	private void updateAttributes(IGraph graph, Map<INode, INode.Attribute> va, final Map<IEdge, IEdge.Attribute> ea) {
		for (INode v : graph.getNodes()) {
			va.put(v, v.getAttribute());
			for (IEdge e : graph.getOutEdges(v)) {
				ea.put(e, e.getAttribute());
			}
		}
	}

	private void checkAttributePreservation(IGraph graph, Map<INode, INode.Attribute> va, final Map<IEdge, IEdge.Attribute> ea) {
		for (INode v : graph.getNodes()) {
			assertEquals(String.format("[Step %d, Node %d] -", step, v.getId()), va.get(v), v.getAttribute());
			for (IEdge e : graph.getOutEdges(v)) {
				assertEquals(String.format("[Step %d, Edge %d] -", step, e.getId()), ea.get(e), e.getAttribute());
			}
		}
	}
	
	private void checkAttributes(IGraph graph) {
		
		for (INode v : graph.getNodes()) {
			
			switch (v.getAttribute()) {
				case CURRENTLY_SETTLED:
					for (INode u : graph.getNodes()) {
						if (u.getAttribute() != INode.Attribute.START_NODE && u.getAttribute() != INode.Attribute.SETTLED)
							assertTrue(v.getDistance() <= u.getDistance());
					}
					break;
				case NOT_VISITED:
					assertTrue(v.getDistance() == Double.POSITIVE_INFINITY);
					break;
				case PATH_FOUND:
					assertTrue(v.getDistance() < Double.POSITIVE_INFINITY);
					assertEquals(Double.POSITIVE_INFINITY, c.getOldDistance(v));
					break;
				case PATH_FOUND_NEXT_SETTLED:
					assertTrue(v.getDistance() < Double.POSITIVE_INFINITY);
					assertEquals(Double.POSITIVE_INFINITY, c.getOldDistance(v));					
					for (INode u : graph.getNodes()) {
						if (u.getAttribute() != INode.Attribute.START_NODE && u.getAttribute() != INode.Attribute.SETTLED && u.getAttribute() != INode.Attribute.CURRENTLY_SETTLED)
							assertTrue(String.format("%s %s", v, u), v.getDistance() <= u.getDistance());
					}
					break;
				case PATH_IMPROVED:
					assertTrue(v.getDistance() < Double.POSITIVE_INFINITY);
					assertTrue(c.getOldDistance(v) < Double.POSITIVE_INFINITY);
					break;
				case PATH_IMPROVED_NEXT_SETTLED:
					assertTrue(v.getDistance() < Double.POSITIVE_INFINITY);
					assertTrue(c.getOldDistance(v) < Double.POSITIVE_INFINITY);
					for (INode u : graph.getNodes()) {
						if (u.getAttribute() != INode.Attribute.START_NODE && u.getAttribute() != INode.Attribute.SETTLED && u.getAttribute() != INode.Attribute.CURRENTLY_SETTLED)
							assertTrue(v.getDistance() <= u.getDistance());
					}
					break;
				case SETTLED:
					for (INode u : graph.getNodes()) {
						if (u.getAttribute() != INode.Attribute.START_NODE && u.getAttribute() != INode.Attribute.SETTLED)
							assertTrue(v.getDistance() <= u.getDistance());
					}
					break;
				case START_NODE:
					assertEquals(0d, v.getDistance());
					break;
				case VISITED:
					assertTrue(v.getDistance() < Double.POSITIVE_INFINITY);
					break;
				case VISITED_NEXT_SETTLED:
					assertTrue(v.getDistance() < Double.POSITIVE_INFINITY);
					if (c.getChangedNodes().contains(v)) assertEquals(v.getDistance(), c.getOldDistance(v));
					for (INode u : graph.getNodes()) {
						if (u.getAttribute() != INode.Attribute.START_NODE && u.getAttribute() != INode.Attribute.SETTLED && u.getAttribute() != INode.Attribute.CURRENTLY_SETTLED)
							assertTrue(v.getDistance() <= u.getDistance());
					}
					break;
			}

			for (IEdge e : graph.getOutEdges(v)) {
				switch (e.getAttribute()) {
					case ADDED_TO_SHORTEST_PATH:
						assertNotSame(IEdge.Attribute.ON_SHORTEST_PATH, c.getOldAttribute(e));
						assertNotSame(IEdge.Attribute.ADDED_TO_SHORTEST_PATH, c.getOldAttribute(e));
						break;
					case NOT_VISITED:
						break;
					case ON_SHORTEST_PATH:
						break;
					case REMOVED_FROM_SHORTEST_PATH:
						assertTrue(c.getOldAttribute(e) == IEdge.Attribute.ON_SHORTEST_PATH || c.getOldAttribute(e) == IEdge.Attribute.ADDED_TO_SHORTEST_PATH);
						break;
					case VISITED:
						break;
				}
			}
		}
		
		
		
	}

	// ####################################################################################

	private void printMemoryStats() {
		final int kilobytes = Math.round((Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory()) / 1024f);
		System.out.printf("%.3f MB of %.3f MB free\n", kilobytes / 1024f, Runtime.getRuntime().maxMemory() / (1024f * 1024f));
	}

	@Test
	public void testBigGraphWithTiming() {
		long start, end;
		MyNodeFactory.getInstance().reset();
		MyEdgeFactory.getInstance().reset();
		final MyPseudoPlanarGraphGenerator<INode, IEdge> gen = new MyPseudoPlanarGraphGenerator<INode, IEdge>(MyGraphFactory.getInstance(), MyNodeFactory.getInstance(),
				MyEdgeFactory.getInstance());

		printMemoryStats();
		start = System.currentTimeMillis();
		final IGraph graph = (IGraph) gen.generateGraph(100, 100, 1, 0);
		end = System.currentTimeMillis();
		System.out.printf("Generate Graph: %.2fs\n", (end - start) / (1000d));
		printMemoryStats();
		System.out.println();

		final DijkstraShortestPath algo = new DijkstraShortestPath();
		algo.graphReplaced(graph, null);

		printMemoryStats();
		start = System.currentTimeMillis();
		algo.initialize(graph.getNodes().iterator().next());
		end = System.currentTimeMillis();
		System.out.printf("Initialize Graph: %.2fs\n", (end - start) / (1000d));
		printMemoryStats();
		System.out.println();

		printMemoryStats();
		start = System.currentTimeMillis();
		algo.run();
		end = System.currentTimeMillis();
		System.out.printf("Run Algorithm: %.2fs\n", (end - start) / (1000d));
		printMemoryStats();
	}

}
